@@ -3,6 +3,8 @@
 from __future__ import print_function
 import sys
 import os
+import logging
+import os.path as op
 import collections
 import re
 import itertools as it
@@ -10,10 +12,18 @@ from collections import OrderedDict
 import networkx as nx
 import numpy as np
 from heapq import *
+import random
 
 REQUIRED = ['family_id', 'sample_id', 'paternal_id',
             'maternal_id', 'sex', 'phenotype']
 
+log = logging.getLogger(__name__)
+
+try:
+    zip = it.izip
+except AttributeError:
+    pass
+
 try:
     import matplotlib
     matplotlib.use("Agg")
@@ -67,7 +77,7 @@ def lowest_common_ancestors(G, targets):
             n = q.popleft()
             if n in parents:
                 continue
-            for p in G.successors_iter(n):
+            for p in G.successors(n):
                 q.append(p)
             parents.add(n)
 
@@ -81,13 +91,13 @@ def lowest_common_ancestors(G, targets):
     lowest_common = set()
     if common_ancestors is not None:
         for p in common_ancestors:
-            if any(child not in common_ancestors and child in all_ancestors for child in G.predecessors_iter(p)):
+            if any(child not in common_ancestors and child in all_ancestors for child in G.predecessors(p)):
                 lowest_common.add(p)
 
     return lowest_common
 
 if sys.version_info[0] == 3:
-    basestring = str
+    basestring = (str, bytes)
 
 class UNKNOWN(object):
     def __str__(self):
@@ -148,16 +158,20 @@ class Sample(object):
         self.header[:6] = REQUIRED
         self.attrs = extra_attrs or []
 
-    def dict(self):
-        d = dict((k, getattr(self, k)) for k in self.header)
+    def dict(self, exclude=None):
+        if exclude is None:
+            exclude = {}
+        d = OrderedDict((k, getattr(self, k)) for k in self.header if not k in exclude)
         for k in ('maternal_id', 'paternal_id', 'sex'):
             d[k] = str(d[k])
         d['phenotype'] = 'affected' if self.affected else 'unaffected'
         return d
 
     def __eq__(self, other):
-        if self is None or other is None:
+        if self is None or other is None or other is UNKNOWN:
             return False
+        if isinstance(other, basestring):
+            return self.sample_id == other
         return (self.sample_id == other.sample_id) and (self.family_id ==
                                                         other.family_id)
     def __hash__(self):
@@ -211,18 +225,6 @@ class Sample(object):
 
     dad = property(_get_dad, _set_dad)
 
-    def __repr__(self):
-        v = "%s('%s', '%s', '%s', '%s', '%s', '%s'" % (self.__class__.__name__,
-                                                       self.family_id,
-                                                       self.sample_id,
-                                                       self.paternal_id,
-                                                       self.maternal_id,
-                                                       SEX.rlookup(self.sex),
-                                                       PHENOTYPE.rlookup(self.affected))
-        if self.attrs:
-            v += ", " + str(self.attrs)
-        v += ")"
-        return v
 
     def __getattr__(self, key):
         if not key in self.header:
@@ -260,6 +262,18 @@ class Sample(object):
         return cls(row[0], row[1], row[2] or "-9", row[3] or "-9", row[4], row[5],
                    row[6:] if len(row) > 6 else None, header=header, warn=warn)
 
+    def __repr__(self):
+        v = "%s('%s', '%s', '%s', '%s', '%s', '%s'" % (self.__class__.__name__,
+                                                       self.family_id,
+                                                       self.sample_id,
+                                                       self.paternal_id,
+                                                       self.maternal_id,
+                                                       SEX.rlookup(self.sex),
+                                                       PHENOTYPE.rlookup(self.affected))
+        if self.attrs:
+            v += ", " + str(self.attrs)
+        v += ")"
+        return v
     def __str__(self):
         v = "%s %s %s %s %s %s" % (self.family_id, self.sample_id,
                                    self.paternal_id,
@@ -279,14 +293,14 @@ class Family(object):
     """
     def __init__(self, samples, warn=True):
         assert len(set(s.family_id for s in samples)) == 1
-        self.unknown_samples = []
+        self.unknown_samples = {}
         self.samples = samples
         self.warn = warn
         self._build()
         for u in self.unknown_samples:
             if self.warn:
-                print("unknown sample: %s in family: %s" % (u,
-                      samples[0].family_id), file=sys.stderr)
+                log.info("unknown sample: %s in family: %s" % (u,
+                      samples[0].family_id))
 
     def __iter__(self):
         self._index = 0
@@ -350,17 +364,22 @@ class Family(object):
             if s.paternal_id != UNKNOWN:
                 try:
                     s.dad = by_id[s.paternal_id]
-                    s.dad.kids.append(s)
                 except KeyError:
-                    s.dad = None
-                    self.unknown_samples.append(s.paternal_id)
+                    if not s.paternal_id in self.unknown_samples:
+                        self.unknown_samples[s.paternal_id] = Sample(s.family_id, s.paternal_id, "-9", "-9", "1", "-9")
+                    #self.unknown_samples[s.paternal_id].kids.append(s)
+                    s.dad = self.unknown_samples[s.paternal_id]
+                s.dad.kids.append(s)
+
             if s.maternal_id != UNKNOWN:
                 try:
                     s.mom = by_id[s.maternal_id]
-                    s.mom.kids.append(s)
                 except KeyError:
-                    s.mom = None
-                    self.unknown_samples.append(s.maternal_id)
+                    if not s.maternal_id in self.unknown_samples:
+                        self.unknown_samples[s.maternal_id] = Sample(s.family_id, s.maternal_id, "-9", "-9", "2", "-9")
+                    #self.unknown_samples[s.maternal_id].kids.append(s)
+                    s.mom = self.unknown_samples[s.maternal_id]
+                s.mom.kids.append(s)
 
     @property
     def affecteds(self):
@@ -421,6 +440,7 @@ class Ped(object):
         self.warn = warn
         self._parse(ped)
         self._graph = None
+        self._cache = {}
 
     def _parse(self, fh):
         self.header = None
@@ -432,7 +452,7 @@ class Ped(object):
                 toks = [x.strip() for x in re.split("\s+", l.strip("\n"))]
             else:
                 toks = l.split(sep)
-            if i == 0 and (toks[0][0] == "#" or toks[0] == "family_id"):
+            if i == 0 and (toks[0][0] == "#" or toks[0].lower() in ("family_id", "kindred_id")):
                 if toks[0][0] == "#":
                     toks[0] = toks[0][1:]
                 self.header = toks
@@ -474,12 +494,14 @@ class Ped(object):
     def __repr__(self):
         return "%s('%s')" % (self.__class__.__name__, self.filename)
 
-    def to_json(self, samples=None):
+    def to_json(self, samples=None, exclude=None):
         import json
+        if exclude is None: exclude = {}
         if samples is None:
-            return json.dumps([s.dict() for s in self.samples()])
+            dicts = [s.dict(exclude=exclude) for s in self.samples()]
         else:
-            return json.dumps([s.dict() for s in self.samples() if s.sample_id in set(samples)])
+            dicts = [s.dict(exclude=exclude) for s in self.samples() if s.sample_id in set(samples)]
+        return json.dumps(dicts)
 
     def relation(self, sample_a, sample_b):
         if isinstance(sample_a, basestring):
@@ -496,19 +518,36 @@ class Ped(object):
             b = sample_b
 
         if a is None or b is None:
-            return 'unknown'
+            #    return 'unknown'
+            a = a or sample_a
+            b = b or sample_b
 
-        # TODO: should we check anyway or just bail early like this
-        if a.family_id != b.family_id:
-            return 'unrelated'
+        else:
+            if a.family_id != b.family_id:
+                return 'unrelated'
 
         if set(a.kids).intersection(b.kids):
             return "mom-dad"
-
         if a.mom == b or b.mom == a or a.dad == b or b.dad == a:
             return 'parent-child'
 
-        # TODO: do BFS
+        # grand-parents
+        for (s, o) in ((a, b), (b, a)):
+            for aa in ('mom', 'dad'):
+                parent = getattr(s, aa, None)
+                if parent is None: continue
+                for bb in ('mom', 'dad'):
+                    gparent = getattr(parent, bb, None)
+                    if gparent is None: continue
+                    if gparent == o: return "grandchild"
+                    if o in gparent.kids:
+                        return "niece/nephew"
+                    for cc in ('mom', 'dad'):
+                        ggparent = getattr(gparent, cc, None)
+                        if ggparent is None: continue
+                        if ggparent == o: return "great-grandchild"
+
+
         if a.mom is None and a.dad is None and b.mom is None and b.dad is None:
             return 'unrelated'
 
@@ -518,39 +557,63 @@ class Ped(object):
         if (a.mom is not None and a.mom == b.mom) or (a.dad is not None and a.dad == b.dad):
             return 'siblings'
 
-        else:
-            return 'related level 2'
+        if None in (a.mom, b.mom, a.dad, b.dad):
+            return 'related at unknown level'
 
-    def get(self, sample_id, family_id=None, cache={}):
+
+        if 'siblings' in self.relation(a.mom, b.mom):
+            return 'cousins'
+        if 'siblings' in self.relation(a.dad, b.dad):
+            return 'cousins'
+        if 'siblings' in self.relation(a.mom, b.dad):
+            return 'cousins'
+        if 'siblings' in self.relation(a.dad, b.mom):
+            return 'cousins'
+
+
+        return 'unknown'
+
+    def get(self, sample_id, family_id=None, samples=None):
         """
         get takes a sample id and optional family_id and returns the object(s)
         associated with it.
         """
-        if (sample_id, family_id) in cache:
-            return cache[(sample_id, family_id)]
-        a = [x for x in self.samples() if x.sample_id == sample_id]
+        if (sample_id, family_id) in self._cache:
+            return self._cache[(sample_id, family_id)]
+        a = [x for x in (samples or self.samples()) if x.sample_id == sample_id]
         if len(a) > 1 and family_id is None:
-            print("multiple samples found in ped file for %s" % sample_id, file=sys.stderr)
+            log.info("multiple samples found in ped file for %s" % sample_id)
 
         if family_id is not None:
             a = [x for x in a if x.family_id == family_id]
         if len(a) > 1:
-            print("multiple samples found in ped file for %s" % sample_id, file=sys.stderr)
+            log.info("multiple samples found in ped file for %s" % sample_id)
         elif len(a) == 0:
-            print("no sample found in ped file for %s" % sample_id, file=sys.stderr)
+            for f in self.families.values():
+                if sample_id in f.unknown_samples:
+                    a = f.unknown_samples[sample_id]
+                    return a
+            else:
+                log.info("no sample found in ped file for %s" % sample_id)
             return None
         elif len(a) == 1:
             a = a[0]
-        cache[(sample_id, family_id)] = a
+        self._cache[(sample_id, family_id)] = a
         return a
 
     def _setup_graph(self):
         gr = self._graph = nx.DiGraph()
         for s in self.samples():
             if s.mom is not None:
-                gr.add_edge(s.sample_id, s.mom.sample_id)
+                if isinstance(s.mom, basestring):
+                    gr.add_edge(s.sample_id, s.mom)
+                elif s.mom != UNKNOWN:
+                    gr.add_edge(s.sample_id, s.mom.sample_id)
             if s.dad is not None:
-                gr.add_edge(s.sample_id, s.dad.sample_id)
+                if isinstance(s.dad, basestring):
+                    gr.add_edge(s.sample_id, s.dad)
+                elif s.dad != UNKNOWN:
+                    gr.add_edge(s.sample_id, s.dad.sample_id)
             #if s.dad is None and s.mom is None:
                 # shoudl top-node be s.family_id?
             #    gr.add_edge(s.sample_id, None)
@@ -599,10 +662,11 @@ class Ped(object):
         path = a_paths + b_paths
         return n * 2.0**-len(path)
 
-    def sex_check(self, vcf_path, min_depth=6,
+    def sex_check(self, vcf_path, min_depth=7,
                   skip_missing=True,
                   plot=False,
-                  cutoff=0.5,
+                  cutoff=0.6,
+                  n_sites=100000,
                   pars=('X:10000-2781479', 'X:155701382-156030895')):
         """
         Check that the sex reported in the ped file matches that inferred
@@ -611,8 +675,9 @@ class Ped(object):
 
         :param vcf str:  path to vcf
         :param min_depth int: minimum depth of variants to consider (in at least 50% of samples).
+        :param n_sites int: stop after sampling this many sites in the X chromosome.
         :param skip_missing bool: don't consider samples that are not in the ped file
-        :param plot bool: render a plot of the distributions by gender.
+        :param plot bool: render a plot of the distributions by sex.
         :param pars tuple(str): pseudo autosmal regions
 
         :return: pandas.DataFrame
@@ -626,60 +691,82 @@ class Ped(object):
                   samples=[s.sample_id for s in self.samples()])
 
         pars = [x.split(':') for x in pars]
-        pars = [(x[0], map(int, x[1].split('-'))) for x in pars]
+        pars = [(x[0], [int(p) for p in x[1].split('-')]) for x in pars]
 
         chrom = pars[0][0]
+        try:
+            next(vcf(chrom))
+        except StopIteration:  # check for chrX
+            if chrom.startswith('chr'): raise
+            chrom = 'chr' + chrom
+            # try with chr prefix
+            try:
+                next(vcf(chrom))
+            except StopIteration:
+                log.warning("no values found for sex chromosome")
+                return None
+            pars = [('chr' + c, v) for c, v in pars]
+
 
         hom_ref = np.zeros(len(vcf.samples), dtype=int)
         het     = np.zeros(len(vcf.samples), dtype=int)
         hom_alt = np.zeros(len(vcf.samples), dtype=int)
         kept, skipped = 0, 0
         for variant in vcf(chrom):
-            depth_filter = variant.gt_depths >= min_depth
-            gt_types = variant.gt_types
-            if any(s <= variant.end and e >= variant.end for chrom, (s, e) in pars):
+            # skip multi-nucleotide
+            if len(variant.REF) > 1: continue
+            # skip multiple alternates
+            if len(variant.ALT) > 1: continue
+            if variant.call_rate < 0.5: continue
+            if variant.aaf < 0.01: continue
+            if any(s <= variant.start and e >= variant.start for chrom, (s, e) in pars):
                 skipped += 1
                 continue
+            depth_filter = variant.gt_depths >= min_depth
+            gt_types = variant.gt_types
+
             hom_ref += (gt_types == 0) & depth_filter
             hom_alt += (gt_types == 2) & depth_filter
             het += (gt_types == 1) & depth_filter
             kept += 1
+            if kept >= n_sites: break
 
         # this should be high for females and low for males
         het_ratio = het.astype(float) / (hom_alt)
-        print("sex-check: %s skipped / %d kept" % (skipped, kept), file=sys.stderr)
+        log.info("sex-check: %s skipped / %d kept" % (skipped, kept))
 
         plot_vals = {'male': [], 'female': [], 'male_errors': [],
-                'female_errors': [], 'male_samples': [], 'female_samples':[]}
+                    'unknown': [], 'unknown_errors': [], 'unknown_samples': [],
+                    'female_errors': [], 'male_samples': [], 'female_samples':[]}
         res = []
         for i, s in enumerate(vcf.samples):
             try:
-                ped_sex = self[s].sex
+                ped_sex = str(self[s].sex)
             except KeyError:
                 if skip_missing:
-                    print(s)
+                    log.info(s)
                     continue
                 ped_sex = "NA"
             val = -0.06 if np.isnan(het_ratio[i]) else het_ratio[i]
             predicted_sex = "UNKNOWN" if val < 0 else SEX.MALE if val < cutoff else SEX.FEMALE
             error = predicted_sex != ped_sex
-            if ped_sex == "NA":
-                error = "NA"
-            else:
-                try:
-                    plot_vals[ped_sex].append(val)
-                except KeyError:
-                    # sex unknown
-                    continue
+            if ped_sex in ("NA", "-9"):
+                ped_sex = "unknown"
+            try:
+                plot_vals[ped_sex].append(val)
+            except KeyError:
+                log.info(s, ped_sex, ped_sex.strip() == "-9")
+                # sex unknown
+                raise
 
             plot_vals[ped_sex + '_errors'].append(error)
             plot_vals[ped_sex + '_samples'].append(s)
-            res.append(dict(sample=s, ped_sex=ped_sex, hom_ref_count=hom_ref[i],
+
+            res.append(dict(sample_id=s, ped_sex=ped_sex, hom_ref_count=hom_ref[i],
                        het_count=het[i], hom_alt_count=hom_alt[i],
                        het_ratio=val, predicted_sex=predicted_sex,
                        error=error))
 
-
         if not plot:
             return pd.DataFrame(res)
 
@@ -698,14 +785,16 @@ class Ped(object):
             return colors
 
         fcolors = [colors[1] if e else colors[0] for e in plot_vals['female_errors']]
-        plt.scatter((np.random.uniform(size=len(plot_vals['female'])) - 0.5) / 4.0, plot_vals['female'],
+        female_xs = (np.random.uniform(size=len(plot_vals['female'])) - 0.5) / 2.1
+        plt.scatter(female_xs, plot_vals['female'],
                     c=fcolors,
                     s=s,
                     edgecolors=update_colors(fcolors, plot_vals['female']),
                     marker='o')
 
         mcolors = [colors[0] if e else colors[1] for e in plot_vals['male_errors']]
-        plt.scatter((np.random.uniform(size=len(plot_vals['male'])) - 0.5) / 4.0 + 1.0, plot_vals['male'],
+        male_xs = (np.random.uniform(size=len(plot_vals['male'])) - 0.5) / 2.1 + 1.0
+        plt.scatter(male_xs, plot_vals['male'],
                     c=mcolors,
                     s=s,
                     edgecolors=update_colors(mcolors, plot_vals['male']),
@@ -713,12 +802,12 @@ class Ped(object):
 
         for i, e in enumerate(plot_vals['female_errors']):
             if not e: continue
-            plt.text(0, plot_vals['female'][i], plot_vals['female_samples'][i],
+            plt.text(female_xs[i], plot_vals['female'][i], plot_vals['female_samples'][i],
                      color=colors[1], fontsize=7)
 
         for i, e in enumerate(plot_vals['male_errors']):
             if not e: continue
-            plt.text(1, plot_vals['male'][i], plot_vals['male_samples'][i],
+            plt.text(male_xs[i], plot_vals['male'][i], plot_vals['male_samples'][i],
                      color=colors[0], fontsize=7)
 
         import matplotlib.patches as mpatches
@@ -726,84 +815,117 @@ class Ped(object):
         c1 = mpatches.Patch(color=colors[1], label="male")
 
         plt.xticks([0, 1], ['female', 'male'])
-        plt.xlim(-0.15, 1.15)
+        plt.xlim(-0.27, 1.27)
         plt.ylim(ymin=-0.08)
-        plt.xlabel('Gender From Ped')
+        plt.axhline(y=cutoff, color='0.75')
+        plt.xlabel('Sex From Ped')
         plt.ylabel('HET / HOM_ALT [higher is more likely female]')
-        plt.legend(handles=[c0, c1], title="Gender predicted\nfrom genotypes",
+        plt.legend(handles=[c0, c1], title="Sex predicted\nfrom genotypes",
                    loc='upper center')
         plt.savefig(plot)
         plt.close()
         return pd.DataFrame(res)
 
-    def het_check(self, vcf, plot=False, min_depth=8):
+    def het_check(self, vcf_path, plot=False, ncpus=1, min_depth=8,
+                  sites=op.join(op.dirname(__file__), '1kg.sites'),
+                  **kwargs):
+        """
+        kwargs is not used, but added here to allow same args as ped_check
+        """
         import cyvcf2
         import numpy as np
+        if ncpus > 16:
+            ncpus = 16
 
         samps = [x.sample_id for x in self.samples()]
-        vcf = cyvcf2.VCF(vcf, gts012=True, samples=samps)
+        vcf = cyvcf2.VCF(vcf_path, gts012=True, samples=samps)
         if sorted(vcf.samples) != sorted(samps):
-            print("warning: sample overlap issues\n\tin vcf, not in ped: %s\n\tin ped, not in vcf: %s" % (
+            log.warning("sample overlap issues\n\tin vcf, not in ped: %s\n\tin ped, not in vcf: %s" % (
                   ",".join(set(vcf.samples) - set(samps)),
-                  ",".join(set(samps) - set(vcf.samples))), file=sys.stderr)
+                  ",".join(set(samps) - set(vcf.samples))))
         if set(vcf.samples) - set(samps) == set(vcf.samples):
             raise Exception("error: no samples from VCF found in ped")
 
-        sample_ranges = vcf.het_check(min_depth=min_depth)
+        samps = vcf.samples
+        sample_ranges, sites, gt_types = cyvcf2.par_het(vcf_path, samps, ncpus,
+                sites, min_depth=min_depth)
+
+        call_rate = (gt_types != 3).mean(axis=1)
+
+        from .pca import pca
+        if plot:
+            pca_plot = plot.replace('het_', 'pca_').replace('het-', 'pca-')
+            if pca_plot == plot:
+                pca_plot, ext = pca_plot.rsplit(".", 1)
+                pca_plot = "%s.%s%s" % (pca_plot, "pca.", ext)
+        else:
+            pca_plot = False
+        pca_df, background_pca_df = pca(pca_plot, gt_types, sites)
 
         # not find outliers.
-        ranges = np.array([d['range'] for d in sample_ranges.values()])
+        depth = np.array([v['median_depth'] for v in sample_ranges.values()])
+        #ranges = np.array([d['range'] for d in sample_ranges.values()])
         ratios = np.array([d['het_ratio'] for d in sample_ranges.values()])
-        ranges_outlier = np.zeros_like(ranges).astype(bool)
-        ratios_outlier = np.zeros_like(ranges).astype(bool)
+
+        bot = depth.mean() - 2 * depth.std()
+        # remove outliers and re-calc.
+        bot = depth[depth > bot].mean() - 5 * depth[depth > bot].std()
+        # care less if we have really high samples so make it 5.
+        top = depth.mean() + 8 * depth.std()
+
+        depth_outlier = ((depth < bot) | (depth > top))
+
         for k, v in sample_ranges.items():
             v['sample_id'] = k
 
 
-        for i in range(2):
-            ranges_mean, ranges_std = np.mean(ranges[~ranges_outlier]), np.std(ranges[~ranges_outlier])
-            ratios_mean, ratios_std = np.mean(ratios[~ratios_outlier]), np.std(ratios[~ratios_outlier])
-            M = i + 2.5
-
-            rmin, rmax = ranges_mean - M * ranges_std, ranges_mean + M * ranges_std
-            ranges_outlier = np.array([not rmin <= d <= rmax for d in ranges])
-
-            rmin, rmax = ratios_mean - M * ratios_std, ratios_mean + M * ratios_std
-            ratios_outlier = np.array([not rmin <= d <= rmax for d in ratios])
-
-        for d, range_o, ratio_o in zip(sample_ranges.values(), ranges_outlier,
-                                       ratios_outlier):
-            d['range_outlier'] = range_o
-            d['ratio_outlier'] = ratio_o
+        for d, depth_o in zip(sample_ranges.values(), depth_outlier):
+            d['depth_outlier'] = depth_o
+            d['idr_baf'] = d.pop('range')
 
         import pandas as pd
-        df = pd.DataFrame(sample_ranges.values())
+        if sys.version_info[0] == 2:
+            df = pd.DataFrame(sample_ranges.values())
+        else:
+            df = pd.DataFrame(list(sample_ranges.values()))
         cols = ['sample_id'] + sorted([x for x in df.columns if x != 'sample_id'])
         df = df[cols]
+
+        df.index = df['sample_id']
+        l = {s: i for i, s in enumerate(samps)}
+        df['call_rate'] = [call_rate[l[s]] for s in df.index]
+
+        if pca_df is not None:
+            # merge the 2 dataframes.
+            pca_df.index = samps
+            df = pd.concat((df, pca_df), axis=1)
+
         if not plot:
-            return df
+            return df, background_pca_df
 
         from matplotlib import pyplot as plt
         import seaborn as sns
         colors = sns.color_palette('Set1', 4)
 
-        cs = [colors[int(v['range_outlier'])] for v in sample_ranges.values()]
-        ecs = ['none' if not v['ratio_outlier'] else 'k' for v in sample_ranges.values()]
+        cs = [colors[1 - int(v['depth_outlier'])] for v in sample_ranges.values()]
+        ecs = ['none' for k in sample_ranges]
 
         s = get_s(np.array([v['median_depth'] for v in sample_ranges.values()]))
 
-        plt.scatter(ranges, ratios, c=cs, edgecolors=ecs, s=s)
+        plt.scatter(depth, ratios, c=cs, edgecolors=ecs, s=s)
 
         for k, v in ((k, v) for k, v in sample_ranges.items()
-                   if v['ratio_outlier'] or v['range_outlier']):
-          plt.text(v['range'], v['het_ratio'], k, color=colors[1], fontsize=7)
+                      if v['depth_outlier']):
+          plt.text(v['median_depth'], v['het_ratio'], k, color=colors[1], fontsize=7)
 
-        plt.xlabel('maf range')
+        plt.xlabel('median depth')
         plt.ylabel('proportion het calls')
         plt.savefig(plot)
-        return df
+        return df, background_pca_df
 
-    def ped_check(self, vcf, ncpus=1, plot=False, min_depth=5, each=1, prefix=''):
+    def ped_check(self, vcf, ncpus=1, plot=False, min_depth=5, each=1,
+            prefix='',
+            sites=op.join(op.dirname(__file__), '1kg.sites')):
         """
         Given the current pedigree and a VCF of genotypes, find sample-pairs where
         the relationship reported in the pedigree file do not match those inferred
@@ -819,104 +941,173 @@ class Ped(object):
         import numpy as np
         import pandas as pd
         vcf_str = vcf
+        np.random.seed(42)
 
-        samps = list(self.samples())
+        ped_samples = list(self.samples())
         if isinstance(vcf, basestring):
-            vcf = cyvcf2.VCF(vcf, gts012=True, samples=[x.sample_id for x in samps])
+            vcf = cyvcf2.VCF(vcf, gts012=True, samples=[x.sample_id for x in ped_samples])
 
         d = cyvcf2.par_relatedness(vcf_str,
-                                   [x.sample_id for x in samps],
+                                   [x.sample_id for x in ped_samples],
                                    ncpus,
+                                   sites,
                                    min_depth=min_depth, each=each)
         cols = ['sample_a', 'sample_b']
         cols += [c for c in d if not c in ('sample_a', 'sample_b') and not c.endswith('error')]
         cols += [c for c in d if c.endswith('error')]
-        if len(samps) > 200:
-            print("large dataset: only reporting pedigree checks where in same"
-                  + " family or relationship does not match expected",
-                  file=sys.stderr)
+        if len(ped_samples) > 200:
+            log.info("large dataset: only reporting pedigree checks where in same"
+                  + " family or relationship does not match expected")
         df = pd.DataFrame(d, columns=cols)
-        a_samples = [self.get(a) for a in df.sample_a]
-        assert all(not isinstance(a, list) for a in a_samples)
-        b_samples = [self.get(b) for b in df.sample_b]
-        assert all(not isinstance(b, list) for b in a_samples)
-        df["pedigree_parents"] = np.array([self.relation(a, b) == 'parent-child' for a, b in
-                                           zip(a_samples, b_samples)])
-        df["pedigree_relatedness"] = np.array([self.relatedness_coefficient(a, b) for a, b in
-                                               zip(df.sample_a, df.sample_b)])
-        df["predicted_parents"] = df['ibs0'] < 0.012
-        df["parent_error"] = df['pedigree_parents'] != df['predicted_parents']
-        df["sample_duplication_error"] = (df['ibs0'] < 0.012) & (df['rel'] > 0.75)
+        del d
+        import gc; gc.collect()
 
-        pr = df['pedigree_relatedness'][:]
-        pr[pr < 0] = 0
-        df["rel_difference"] = pr - df['rel']
+
+        # most of these 2 will be false and 0, respectively
+        df['pedigree_parents'] = np.zeros(len(df), dtype=bool)
+        df['pedigree_relatedness'] = np.zeros(len(df), dtype=np.float32)
+
+        fam_lookup = {s.sample_id: s.family_id for s in ped_samples}
+        same_fam = np.array([fam_lookup[a] == fam_lookup[b] for a, b in zip(df.sample_a, df.sample_b)])
+
+        asample_ids, bsample_ids = df.sample_a[same_fam], df.sample_b[same_fam]
+        idxs, = np.where(same_fam)
+
+        import array
+        parent_is = array.array('L', [])
+        rels = array.array('f', [])
+
+        # if they aren't in the same fam, cant be related.
+        for (i, aid, bid) in zip(idxs, asample_ids, bsample_ids):
+
+            a_sample, b_sample = self.get(aid, samples=ped_samples), self.get(bid, samples=ped_samples)
+            if b_sample in (a_sample.mom, a_sample.dad) or a_sample in (b_sample.mom, b_sample.dad):
+                parent_is.append(i)
+                #df.loc[i, 'pedigree_parents'] = True
+
+            # setting directly is expensive, but we expect that for big cohorts, the above 2 sets will
+            # be relatively rare (and they default to False). Since we need to set the relatedness fcolors
+            # every sample, we use an array.array and then set at the end.
+            #df.loc[i, 'pedigree_relatedness'] = self.relatedness_coefficient(aid, bid)
+            rels.append(max(0, self.relatedness_coefficient(aid, bid)))
+
+        df.loc[np.asarray(parent_is), 'pedigree_parents'] = True
+        df.loc[idxs, 'pedigree_relatedness'] = np.asarray(rels, np.float32)
+
+
+        df['tmpibs0'] = (df['ibs0'] / df['n'].astype(np.float32)).astype(np.float32)
+        df["predicted_parents"] = df['tmpibs0'] < 0.012
+        df["parent_error"] = df['pedigree_parents'] != df['predicted_parents']
+        df["sample_duplication_error"] = (df['tmpibs0'] < 0.012) & (df['rel'] > 0.75)
+
+        pr = df['pedigree_relatedness']
+        df["rel_difference"] = (pr - df['rel']).astype(np.float32)
         # make the column order a bit more sane.
-        if len(samps) > 200:
+        if len(ped_samples) > 200:
             rd = np.abs(df['rel_difference']) > 0.17
 
-            sampling_rate = 1 / (len(samps)**0.6)
-            print(sampling_rate)
+            sampling_rate = 1 / (len(ped_samples)**0.6)
             ru = (np.random.uniform(size=df.shape[0]) < sampling_rate)
-            keep = df.eval('parent_error | sample_duplication_error | predicted_parents| @rd | @ru' +
-                    '| (rel > 0.17) | (ibs0 < 0.02) | (pedigree_relatedness > 0)')
+            df['keep'] = df.eval('parent_error | sample_duplication_error | predicted_parents| @rd | @ru' +
+                    '| (rel > 0.17) | (tmpibs0 < 0.04) | (pedigree_relatedness > 0)')
 
-            same_fam = []
-            for a, b in it.izip(a_samples, b_samples):
-                same_fam.append(a.family_id == b.family_id)
-            keep |= np.array(same_fam, dtype=bool)
+            df['keep'] |= same_fam
 
-        else:
-            keep = np.ones(df.shape[0]).astype(bool)
 
         if not plot:
+            df.drop('tmpibs0', axis=1, inplace=True)
             return df
+
+
+        def asum(a):
+            return np.abs(a).sum()
+            a = np.abs(a)
+            return a[a > 0.08].sum()
+
         from matplotlib import pyplot as plt
         plt.close()
         import seaborn as sns
         sns.set_style('whitegrid')
 
+        # get total rel_difference by sample. large values indicate the likely problem
+        #df['rel_difference'][df['rel_difference'] == 0] = 0.001
+        df.loc[df['rel_difference'] == 0, 'rel_difference'] == 0.001
+
+        sub = df.eval('((rel > 0.1) & (pedigree_relatedness < 0.05)) | ((rel < 0.05) & (pedigree_relatedness > 0.1)) | (rel_difference > 0.1) | (rel_difference < -0.1)')
+        da = df[sub].groupby('sample_a')['rel_difference'].agg(asum)
+        db = df[sub].groupby('sample_b')['rel_difference'].agg(asum)
+        diff = da.add(db, fill_value=0)
+        diff.sort_values(inplace=True, ascending=False)
+
+        diff.to_csv(plot.rsplit(".", 1)[0] + ".rel-difference.csv", index=True,
+                index_label="sample", header=True)
+        del diff; del da; del db; del sub
+
+
         colors = [(0.85, 0.85, 0.85)] + sns.color_palette('Set1', len(set(df['pedigree_relatedness'])))
         n = df['n'] / df['n'].mean()
+        if len(df) < 100:
+            colors[0] = (0.3, 0.3, 0.3)
+
+        mult = 24 if len(df) < 50 else 12
+
+        fig, axesb = plt.subplots(2, 2, figsize=(12, 12))
+        df['tmpibs2'] = df['ibs2'] / df['n'].astype(float)
+        axes = axesb[0]
+        log.info("plotting")
+
+        for k, key in enumerate(('tmpibs0', 'tmpibs2')):
+
+            for i, rc in enumerate(sorted(set(df['pedigree_relatedness']))):
+                sel = df['pedigree_relatedness'] == rc
+                src = ("%.3f" % rc).rstrip('0')
+                # outline parent kid relationships
+                #ec = ['k' if p else 'none' for p in df['pedigree_parents'][sel]]
+                ec = np.where(df.loc[sel, 'pedigree_parents'], 'k', 'none')
+
+                axes[k].scatter(df.loc[sel, 'rel'], df.loc[sel, key],
+                        c=colors[i], linewidth=1, edgecolors=ec,
+                        s=((mult * (i > 0 or len(df) < 36)) + mult * n[sel]),
+                        zorder=i,
+                        alpha=0.80,
+                        label="ped coef: %s" % src)
+                axes[k].set_xlabel('coefficient of relatedness')
+                axes[k].set_ylabel(key[3:])
 
-        for i, rc in enumerate(sorted(set(df['pedigree_relatedness']))):
-            sel = df['pedigree_relatedness'] == rc
-            src = ("%.3f" % rc).rstrip('0')
-            # outline parent kid relationships
-            ec = ['k' if p else 'none' for p in df['pedigree_parents'][sel]]
-            plt.scatter(df['rel'][sel], df['ibs0'][sel],
-                    c=colors[i], linewidth=1, edgecolors=ec,
-                    s=14 * n[sel],
-                    #s=14,
-                    alpha=0.80,
-                    label="ped coef: %s" % src)
-        plt.xlabel('coefficient of relatedness')
-        plt.ylabel('ibs0')
         if prefix:
-            plt.title(prefix)
-        if os.environ.get('FIXED'):
-            print("fixed axes")
-            plt.xlim(-0.2, 1.1)
-            plt.ylim(-0.02, 0.14)
-        plt.legend()
-        xmin, xmax = plt.xlim()
+            fig.suptitle(prefix)
+        xmin, xmax = axes[0].get_xlim()
         if xmin < -0.3:
-            plt.xlim(xmin=-0.3)
+            axes[0].set_xlim(xmin=-0.3)
         if xmax > 1.25:
-            plt.xlim(xmax=1.25)
+            axes[0].set_xlim(xmax=1.25)
+        df.drop(['tmpibs0', 'tmpibs2'], axis=1, inplace=True)
 
-        ymin, ymax = plt.ylim()
-        if ymin < -0.02:
-            plt.ylim(ymin=-0.02)
-        if ymax > 0.20:
-            plt.ylim(ymax=0.20)
+        axes = axesb[1]
+        for k, key in enumerate(('ibs2', 'shared_hets')):
+            for i, rc in enumerate(sorted(set(df['pedigree_relatedness']))):
+                    sel = df['pedigree_relatedness'] == rc
+                    src = ("%.3f" % rc).rstrip('0')
+                    # outline parent kid relationships
+                    #ec = ['k' if p else 'none' for p in df['pedigree_parents'][sel]]
+                    ec = np.where(df.loc[sel, 'pedigree_parents'], 'k', 'none')
+                    axes[k].scatter(df.loc[sel, 'ibs0'], df.loc[sel, key],
+                            c=colors[i], linewidth=1, edgecolors=ec,
+                            s=((mult * (i > 0 or len(df) < 36)) + mult * n[sel]),
+                            zorder=i,
+                            alpha=0.80,
+                            label="ped coef: %s" % src)
+                    axes[k].set_xlabel('ibs0')
+                    axes[k].set_ylabel(key)
+
+        plt.legend()
 
         if plot is True:
             plt.show()
         else:
             plt.savefig(plot)
         plt.close()
-        return df.ix[keep, :]
+        return df
 
     def summary(self):
         atrios, aquads = 0, 0
